=head1 QualifierEditor::HitInfo

QualifierEditor::App - This package contains the guts of the qualifier_editor
application.

=cut

package QualifierEditor::App;

use strict;
use Exporter;

use QualifierEditor::HitInfo;
use QualifierEditor::HitInfoCollection;

use vars qw(@ISA @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT_OK = qw (create_app);

my $global_font           = "fixed";
my $id_desc_width         = 110;
my $button_xpad           = 1;
my $button_ypad           = 1;
my $qualifier_t_height    = 10;
my $fasta_t_height        = 20;
my $fasta_button_t_height = 12;
my $srs_command           = "getz";
my $pointer_cursor        = "top_left_arrow";
my $button_pack_padx      = 5;

my $NETSCAPE = "/usr/local/lib/communicator/real-netscape";

# change this variable to list the databases to search for the IDs
my $DATABASES = "embl emblnew ensembl swissprot swissnew pir sptrembl remtrembl tremblnew";

$DATABASES =~ s/ /%20/g;

# change this to point to the wgetz script of your SRS server
my $SRS_SERVER = "www.sanger.ac.uk/srs6bin/cgi-bin/wgetz?-e+";


# change this subroutine to change the generated qualifiers
sub get_new_qualifiers ($$$)
{
  my ($old_qualifiers_string, $info_holder, $note_maker_sub) = @_;

  my @editor_info_refs = $info_holder->all_values ();

  my $orth_ref = get_orth_ref ($info_holder);

  my $new_qualifiers = "";

  my @selected_info_refs;

  if (defined $orth_ref) {
    push @selected_info_refs, $orth_ref;
  }

  push @selected_info_refs, get_para_refs ($info_holder);

  if (!@selected_info_refs) {
    return $old_qualifiers_string;
  }

  my $note_value =
    "Similar to " . ($note_maker_sub->(shift @selected_info_refs));

  for my $info_ref (@selected_info_refs) {
    $note_value .= ", and to " . $note_maker_sub-> ($info_ref);
  }

  $new_qualifiers .= qq|/note="$note_value"\n|;

  if (defined $orth_ref) {
    my $orth_id = $orth_ref->{id};

    my $orth_product = $orth_ref->get_product ();
    my $orth_ecnumber = $orth_ref->get_ecnumber ();
    my $orth_gene_name = $orth_ref->get_gene_name ();

    if (defined $orth_product) {
      $new_qualifiers .= qq|/product="$orth_product"\n|;
    }
    if (defined $orth_ecnumber) {
      $new_qualifiers .= qq|/EC_number="$orth_ecnumber"\n|;
    }
    if (defined $orth_gene_name) {
      $new_qualifiers .= qq|/gene="$orth_gene_name"\n|;
    }
  }

  return $new_qualifiers . $old_qualifiers_string;

}

######################################################################



# returns the contents of the file given by the /fasta_file or
# /blast_file in the given qualifiers
sub get_results_file ($$$)
{
  my ($base_directory, $qualifier_string, $qualifier_name) = @_;

#  print STDERR "$base_directory, $qualifier_string, $qualifier_name\n";

  my $results_file_name;

  if ($qualifier_string =~ m!/$qualifier_name="([^\"]+)"!) {
    $results_file_name = $1;
  } else {
    die "can't find /$qualifier_name qualifier\n";
  }

  my @return_lines = ();

  open IN_FILE, "$base_directory/$results_file_name"
     or die "can't open $base_directory/$results_file_name\n";

  my $line;

  while (defined ($line = <IN_FILE>)) {
    push @return_lines, $line;
  }

  close IN_FILE;

  return @return_lines;
}

# Create a new Scrolled Text widget, parse the given fasta results
# and put the results in the new widget.  The sequence ID at the start
# of each alignment will be tagged with the id of the hit.

# Returns ($new_text_widget, $info_holder)
sub make_fasta_text ($\@)
{
  my ($parent, $fasta_file_text_ref) = @_;

  my @fasta_file_text = @{$fasta_file_text_ref};

  my $fasta_text_widget = $parent->Scrolled ("Text",
                                             -font => $global_font,
                                             -scrollbars => 'ose',
                                             -background => 'white',
                                             -width => 80);

  $fasta_text_widget->pack (-side => "right",
                            -expand => "y", -fill => "both");

  my $info_holder = new QualifierEditor::HitInfoCollection ();

  my %seen_ids = ();

  my $seen_top_of_summary = 0;
  my $seen_bottom_of_summary = 0;

  # the the id on the top of the last alignment line
  my $last_id;

  for my $line (@fasta_file_text) {
    if (!$seen_top_of_summary && $line =~ m/^The best scores are/) {
      $seen_top_of_summary = 1;
    } else {
      if ($seen_top_of_summary && !$seen_bottom_of_summary) {
        if ($line =~ m/^$/) {
          $seen_bottom_of_summary = 1;
        } else {
          # handle one line of the summary
          my ($id, $desc, $initn, $init1, $opt, $z_score, $e_value) =
            ($line =~ m{ ^(\S+)\s+    # id
                          (.*?)\s+     # description
                          ([\.\d]+)\s+ # initn
                          ([\.\d]+)\s+ # init1
                          ([\.\d]+)\s+ # opt
                          ([\.\d]+)\s+ # z-score
                          ([\.\de\-\+]+)\s* # evalue
                          $
                      }xg);

          my $orth_flag = 0;
          my $para_flag = 0;

          if (defined $id) {
            if (!$seen_ids{$id}) {
              my $new_info =
                new QualifierEditor::HitInfo (id            => $id,
                                              desc          => $desc,
                                              initn         => $initn,
                                              init1         => $init1,
                                              opt           => $opt,
                                              z_score       => $z_score,
                                              e_value       => $e_value,
                                              orth_flag_ref => \$orth_flag,
                                              para_flag_ref => \$para_flag,
                                             );

              $info_holder->add ($new_info);

              $seen_ids{$id} = $new_info;
            }
          } else {
            die "can't uderstand: $line\n";
          }
        }
      }
    }

    if ($seen_bottom_of_summary && $line =~ m/^(\S+)(\s+.*\s\((\d+)\s+aa\))/) {
      # this is the start of an alignment section

      $last_id = $1;

      my $rest_of_line = $2;

      exists $seen_ids{$last_id} ||
         die "error while reading the alignments: id $last_id doesn't exist\n";

      my $last_hit_info = $seen_ids{$last_id};

      $last_hit_info->{hit_length} = $3;

      if (!exists $last_hit_info->{_seen_alignment}) {
        # only do this if we haven't seen the alignment for this id
        # before (to handle duplicates)

        $fasta_text_widget->tagConfigure ($last_id, -background => "white");

        $fasta_text_widget->insert ('end', $last_id, $last_id);

        $fasta_text_widget->insert ('end', $rest_of_line . "\n");
      } else {
        $fasta_text_widget->insert ('end', $last_id . $rest_of_line . "\n");
      }
    } else {

      if ($line =~ m{ Smith-Waterman\s+score:\s+
                      (\d+);\s+   # sw_score
                      ([\d\.]+)\%\s+  # percent_id
                      identity\s+in\s+(\d+)\s+aa # overlap
                    }x) {

        exists $seen_ids{$last_id} || die "id $last_id doesn't exist\n";

        if (!exists $seen_ids{$last_id}{_seen_alignment}) {
          # only do this if we haven't seen the alignment for this id
          # before (to handle duplicates)
          my $info_ref = $seen_ids{$last_id};

          $info_ref->{sw_score} = $1;
          $info_ref->{percent_id} = $2;
          $info_ref->{overlap} = $3;
        }

        $seen_ids{$last_id}{_seen_alignment} = 1;
      }

      $fasta_text_widget->insert ('end', $line);
    }
  }

  $fasta_text_widget->configure (-state => 'disabled');

  return ($fasta_text_widget, $info_holder);
}

# Create a new Scrolled Text widget, parse the given blastp results
# and put the results in the new widget.  The sequence ID at the start
# of each alignment will be tagged with the id of the hit.

# Returns ($new_text_widget, $info_holder)
sub make_blastp_text ($\@)
{
  my ($parent, $blastp_file_text_ref) = @_;

  my @blastp_file_text = @{$blastp_file_text_ref};

  my $blastp_text_widget = $parent->Scrolled ("Text",
                                             -font => $global_font,
                                             -scrollbars => 'ose',
                                             -background => 'white',
                                             -width => 80);

  $blastp_text_widget->pack (-side => "right",
                            -expand => "y", -fill => "both");

  my $info_holder = new QualifierEditor::HitInfoCollection ();

  my %seen_ids = ();

  my $seen_top_of_summary = 0;
  my $seen_bottom_of_summary = 0;

  # the the id on the top of the last alignment header section
  my $last_id;

  for my $line (@blastp_file_text) {
    if (!$seen_top_of_summary && $line =~ m/^The best scores are/) {
      $seen_top_of_summary = 1;
    } else {
      if ($seen_top_of_summary == 1) {
        # 2 is a flag that indicates that we have seen the blank
        $seen_top_of_summary = 2;

        if ($line =~ m/^$/) {
          warn "warning: expected a blank line but got: $line\n";
        } else {
          $blastp_text_widget->insert ('end', $line);
          next;
        }
      }

      if ($seen_top_of_summary  && !$seen_bottom_of_summary) {
        if ($line =~ m/^$/) {
          $seen_bottom_of_summary = 1;
        } else {
          # handle one line of the summary
          my ($id, $acc, $desc, $score, $prob, $n_value) =
            ($line =~ m{ ^(\S+)\s+     # id
                          (\S+)\s+     # acc
                          (.*?)\s+     # description
                          ([\.\d]+)\s+ # score
                          ([\.\de\-\+]+)\s+ # prob
                          ([\.\d]+)\s* # n_value
                          $
                      }xg);

          my $orth_flag = 0;
          my $para_flag = 0;

          if (defined $id) {
            if (!$seen_ids{$id}) {
              my $new_info =
                new QualifierEditor::HitInfo (id            => $id,
                                              acc           => $acc,
                                              desc          => $desc,
                                              score         => $score,
                                              prob          => $prob,
                                              n_value       => $n_value,
                                              orth_flag_ref => \$orth_flag,
                                              para_flag_ref => \$para_flag,
                                             );

              $info_holder->add ($new_info);

              $seen_ids{$id} = $new_info;
            }
          } else {
            die "can't uderstand: $line\n";
          }
        }
      }
    }

    if ($seen_bottom_of_summary) {
      if ($line =~ m/^>(\S+)(.*)/) {
        # this is the start of an alignment section

        my $this_id = $1;
        $last_id = $this_id;
        my $rest_of_line = $2;

        if (!exists $seen_ids{$this_id}) {
          warn "error while reading the alignments: " .
               "id $this_id doesn't exist in the summary\n";
        }

        print STDERR "1: $line\n";
        my $last_hit_info = $seen_ids{$this_id};

        $last_hit_info->{hit_length} = $3;

        if (!exists $last_hit_info->{_seen_alignment}) {
          # only do this if we haven't seen the alignment for this id
          # before (to handle duplicates)

          $blastp_text_widget->tagConfigure ($this_id, -background => "white");

          $blastp_text_widget->insert ('end', $this_id, $this_id);

          $blastp_text_widget->insert ('end', $rest_of_line . "\n");
        } else {
          $blastp_text_widget->insert ('end', $this_id . $rest_of_line . "\n");
        }
        next;
      } else {
        # Score = 1922 (676.6 bits), Expect = 1.3e-198, P = 1.3e-198
        if ($line =~ m{^\s*Score\s+=\s+(\d)\s+    # score
                       \((\d+)\s+bits\),\s+         # bits
                       Expect\s+=\s+([\.\de\-\+]+),\s+   # expect
                       P\s+=\s+([\.\de\-\+]+)\s*   # p_value
                       $
                      }x) {
          # second line of the alignment section

          my $info_ref = $seen_ids{$last_id};

          $info_ref->{score} = $1;
          $info_ref->{bits}  = $2;
          $info_ref->{e_value} = $3;
          $info_ref->{p_value} = $1;

            print STDERR "2: $line\n";
        } else {
          # last line of the alignment header
          if ($line =~ m/Identities = \d+\/\d+ ([\d\.]+)%/) {
            print STDERR "3: $line\n";

            my $info_ref = $seen_ids{$last_id};

            $info_ref->{percent_id} = $1;

            $seen_ids{$last_id}{_seen_alignment} = 1;
          }
        }
      }

      $blastp_text_widget->insert ('end', $line);
    }
  }

  $blastp_text_widget->configure (-state => 'disabled');

  return ($blastp_text_widget, $info_holder);
}

sub make_entry_text ($$$$)
{
  my ($info_holder, $this_editor_info, $get_org_flag, $results_type) = @_;

  my $id = $this_editor_info->{id};
  my $desc = $this_editor_info->{desc};
  my $organism = "";
  my $e_value = $this_editor_info->{e_value};
  my $z_score = $this_editor_info->{z_score};
  my $percent_id = $this_editor_info->{percent_id};
  my $overlap = $this_editor_info->{overlap};

  my $max_id_length = $info_holder->max_length ("id");
  my $max_desc_length = $info_holder->max_length ("desc");

  my $max_org_length = 30;

  if ($get_org_flag) {
    $organism = $this_editor_info->get_organism;
  }

  if ($results_type eq "fasta") {
    my $text = (sprintf ("%-${max_id_length}s " .
                         "%-${max_desc_length}s " .
                         "%-${max_org_length}s   %8.8s  %5d  " .
                         "%5.1f", $id, $desc, $organism, $e_value, $z_score,
                         $percent_id) .
                '% id in ' .
                sprintf ("%-8.8s" , sprintf ("%d", $overlap) . 'aa'));

    return $text;
  } else {
    return "blastp-make_entry_text()";
  }
}

sub make_button_row ($$$$$$)
{
  my ($parent, $fasta_t, $this_editor_info, $info_holder,
      $callback, $results_type) = @_;

  my ($callback_sub, @callback_args) = @{$callback};

  my $row_f = $parent->Frame (-relief => "groove", -border => 2);

  # true when the orth button is pressed
  my $orth_flag_ref = $this_editor_info->{orth_flag_ref};

  my $orth_b = $row_f->Checkbutton (-text => "ORTH",
                                    -variable => $orth_flag_ref,
                                    -font => $global_font,
                                    -border => 1,
                                    -relief => "raised",
                                    -padx => $button_xpad,
                                    -pady => $button_ypad);

  $orth_b->pack (-side => "left",
                 -padx => $button_pack_padx);
  $orth_b->configure (-cursor => $pointer_cursor);


  # true when the para button is pressed
  my $para_flag_ref = $this_editor_info->{para_flag_ref};

  my $para_b = $row_f->Checkbutton (-text => "PARA",
                                    -variable => $para_flag_ref,
                                    -font => $global_font,
                                    -border => 1,
                                    -relief => "raised",
                                    -padx => $button_xpad,
                                    -pady => $button_ypad);

  $para_b->pack (-side => "left",
                 -padx => $button_pack_padx);
  $para_b->configure (-cursor => $pointer_cursor);


  my $para_command = sub {
    if (${$para_flag_ref} && ${$orth_flag_ref}) {
      ${$orth_flag_ref} = 0;
    }

    $callback_sub->($info_holder, @callback_args);
  };

  my $orth_command = sub {
    if (${$para_flag_ref} && ${$orth_flag_ref}) {
      ${$para_flag_ref} = 0;
    }

    # if any other orth_flag is set, disable it
    for my $other_info_ref ($info_holder->all_values ()) {
      if (${$other_info_ref->{orth_flag_ref}} &&
          $other_info_ref->{orth_flag_ref} != $orth_flag_ref) {
        ${$other_info_ref->{orth_flag_ref}} = 0;
      }
    }

    $callback_sub->($info_holder, @callback_args);
  };

  $para_b->configure (-command => $para_command);
  $orth_b->configure (-command => $orth_command);


  my $id = $this_editor_info->{id};

  my $text = make_entry_text $info_holder, $this_editor_info, 0, $results_type;

  my $id_desc_l = $row_f->Entry (-textvariable => \$text,
                                 -font => $global_font,
                                 -relief => "flat",
                                 -state => "disabled",
                                 -width => length $text);

  my $delay = 10 + rand 100;

  my $id_desc_l_callback;

  $id_desc_l_callback =
    sub {
      $id_desc_l->updateWidgets;
      if (!$parent->ismapped) {
        $parent->after ($delay, $id_desc_l_callback);
      } else {
        $text = make_entry_text ($info_holder, $this_editor_info, 1,
                                 $results_type);
      }
      $parent->updateWidgets;
    };

  $parent->after ($delay, $id_desc_l_callback);

  $id_desc_l->pack (-side => "left",
                    -padx => $button_pack_padx);



  my $align_command = sub {
    my $start = ($fasta_t->tagRanges ($id))[0];

    $fasta_t->yview ($start);
  };

  my $align_b = $row_f->Button (-text => "ALIGN",
                                -font => $global_font,
                                -padx => $button_xpad,
                                -pady => $button_ypad,
                                -borderwidth => 1,
                                -anchor => 'w',
                                -command => $align_command);

  $align_b->pack (-side => "left",
                  -padx => $button_pack_padx);

  $align_b->configure (-cursor => $pointer_cursor);


  my $srs_command = sub {
    my $url = "http://$SRS_SERVER\[libs=\{$DATABASES\}-ID:$id*\]|[libs-AccNumber:$id*]|[libs-ProteinID:$id*]";

    print "$url\n";

    system "$NETSCAPE", "-remote", "openURL($url)";
  };

  my $srs_b = $row_f->Button (-text => "->SRS",
                              -font => $global_font,
                              -padx => $button_xpad,
                              -pady => $button_ypad,
                              -borderwidth => 1,
                              -anchor => 'w',
                              -command => $srs_command);

  $srs_b->pack (-side => "left",
                -padx => $button_pack_padx);

  $srs_b->configure (-cursor => $pointer_cursor);


#    my $more_l = $row_f->Label (-text => "");

#    my $more_command = sub { $more_l->configure (-text => "foo") };

#    my $more_b = $row_f->Button (-text => "MORE ...",
#                                 -font => $global_font,
#                                 -padx => $button_xpad,
#                                 -pady => $button_ypad,
#                                 -borderwidth => 1,
#                                 -anchor => 'w',
#                                 -command => $more_command);

#    $more_b->pack (-side => "left");


#    $more_l->pack (-side => "left");


  return $row_f;
}

sub get_orth_ref (@)
{
  my ($info_holder) = @_;

  for my $info_ref ($info_holder->all_values) {
    if (${$info_ref->{orth_flag_ref}}) {
      return $info_ref;
    }
  }

  return undef;
}

sub get_para_refs (@)
{
  my ($info_holder) = @_;

  grep {
    if (${$_->{para_flag_ref}}) {
      1;
    } else {
      0;
    }
  } ($info_holder->all_values);
}


sub update_qualifier_t ($$$$)
{
  my ($info_holder, $qualifier_t, $orig_qualifier_string,
      $note_maker_sub) = @_;

  my $new_qualifiers_string =
    get_new_qualifiers ($orig_qualifier_string, $info_holder, $note_maker_sub);

  $qualifier_t->configure (-state => "normal");

  $qualifier_t->delete ('1.0', 'end');
  $qualifier_t->insert ('end', $new_qualifiers_string);

  $qualifier_t->configure (-state => "disabled");
};


# Create and return a new top level window for the QualifierEditor application.
#
# $base_directory - The directory where the fasta/blast result are to be found
# $orig_qualifier_string - A string containing all the qualifiers and values
# $ok_callback - The subroutine to call when the user hits OK
# $cancel_callback - The subroutine to call when the user hits Cancel
# $note_maker_sub - The subroutine to call to generate one part of a
#                   /note qualifier (ie. the text describing one hit).
# $results_type - The type of the results to display, eg. "fasta", "blastp".
#                 This value is used to choose which qualifier to get the
#                 results file from (ie. /fasta_file, /blastp_file etc.) and
#                 to decide how to parse the file.  make_fasta_text() is called
#                 for "fasta", make_blast_text() is called for "blast" and
#                 "blastp".
sub create_app ($$$$$$)
{
  my ($base_directory, $orig_qualifier_string,
      $ok_callback, $cancel_callback, $note_maker_sub, $results_type) = @_;

  my $top = MainWindow->new;

#  $top->appname ("QUED");

#  $top->geometry ("850x750");

  $top->wm ("minsize", 100, 100);

  my $top_frame = $top->Frame ();

  my $qualifier_t = $top_frame->Scrolled ("Text",
                                          -font => $global_font,
                                          -scrollbars => 'ose',
                                          -background => 'white',
                                          -height => $qualifier_t_height,
                                          -width => 80);

  $qualifier_t->pack(-side => "left", -expand => "y", -fill => "both");

  $qualifier_t->insert ('end', $orig_qualifier_string);

# $qualifier_t->configure (-state => "disabled");

  my @results_file_text =
    get_results_file $base_directory, $orig_qualifier_string, "fasta_file";

  my $scrolled_button_text = $top->Scrolled ("Text",
                                             -height => $fasta_button_t_height,
                                             -scrollbars => 'ose');

#    my $button_frame = $scrolled_button_text->Frame ();

#  $scrolled_button_text->windowCreate ('end', -window => $button_frame);
#  $scrolled_button_text->insert ('end', "\n");

  my $results_text_widget;
  my $info_holder;


  if ($results_type eq "fasta") {
    ($results_text_widget, $info_holder) =
      make_fasta_text ($top_frame, @results_file_text);
  } else {
    ($results_text_widget, $info_holder) =
      make_blastp_text ($top_frame, @results_file_text);
  }

  my @editor_info_refs = $info_holder->all_values ();

  # make a row of buttons and label that describe each hit

  for my $this_editor_info (reverse @editor_info_refs) {
    my $new_row = make_button_row ($scrolled_button_text,
                                   $results_text_widget,
                                   $this_editor_info,
                                   $info_holder,
                                   [\&update_qualifier_t,
                                    $qualifier_t,
                                    $orig_qualifier_string,
                                    $note_maker_sub],
                                   $results_type);

    $scrolled_button_text->insert ('1.0', "\n");
    $scrolled_button_text->windowCreate ('1.0', -window => $new_row);

  }

  $scrolled_button_text->configure (-state => "disabled");

  my $first_alignment_tag = $editor_info_refs[0]->{id};

  my $start = $results_text_widget->tagRanges ($first_alignment_tag);

  $results_text_widget->see ($start);


  my $button_f = $top->Frame (-border => 2);

  my $real_ok_callback = sub {
    my ($info_holder) = @_;

    my $new_qualifiers_string =
      get_new_qualifiers ($orig_qualifier_string,
                          $info_holder, $note_maker_sub);

    $ok_callback->($new_qualifiers_string);
  };

  my $ok_b = $button_f->Button (-text => "OK",
                                -font => $global_font,
                                -command => [$real_ok_callback,
                                             $info_holder]);

  my $cancel_b = $button_f->Button (-text => "Cancel",
                                    -font => $global_font,
                                    -command => $cancel_callback);

  $ok_b->pack (-side => "left", -padx => 10, -pady => 3);
  $cancel_b->pack (-side => "left", -padx => 10, -pady => 3);

  $scrolled_button_text->pack(-side => "top",
                              -expand => 0, -fill => "both");

  $top_frame->pack (-side => "top", -expand => "y", -fill => "both");

  $button_f->pack (-side => "top");

  return $top;
}
